package fr.username404.snowygui.gui.feature

import com.mojang.blaze3d.vertex.PoseStack
import fr.username404.snowygui.ClickGui
import fr.username404.snowygui.Snowy
import fr.username404.snowygui.config.Configuration
import fr.username404.snowygui.gui.ColoredElement
import fr.username404.snowygui.utils.FontUtil
import fr.username404.snowygui.gui.feature.ButtonInfo.Companion.Type
import fr.username404.snowygui.utils.RenderingUtil.endDraw
import fr.username404.snowygui.utils.RenderingUtil.prepareDraw
import fr.username404.snowygui.utils.RenderingUtil.colorShader
import kotlin.reflect.full.findAnnotation

sealed class ButtonImpl: ColoredElement(0.0, 0.0, 73, 8, opacity = 0.60F) {
    internal companion object {
        private fun addButtons(vararg buttons: ButtonImpl) {
            buttons.groupBy { impl ->
                ClickGui.clickBoxes.find { box ->
                    with(impl) {
                        if (info.shouldSave() && !info.parent.shouldHide) {
                            Configuration.enabledFeatures[title]?.let { bool ->
                                toggled = bool
                            }
                        }
                        box.isCategory(info.parent)
                    }
                }
            }.entries.forEach { entry ->
                with(entry) {
                    key?.buttons?.clear()
                    value.forEach {
                        if (!it.isDisabled) {
                            key?.buttons?.add(it)
                        }
                    }
                }
            }
        }
        @JvmStatic
        fun initButtons() = addButtons(
            *Snowy.annotatedButtons().mapNotNull {
                try {
                    it.getConstructor().newInstance()
                } catch (_: NoSuchMethodException) {
                    try {
                        it.getDeclaredField("INSTANCE").get(null)
                    } catch (_: NoSuchFieldException) {}
                } as? ButtonImpl
            }.plus(Configuration.foundMacros).toTypedArray()
        )
    }
    var isDisabled: Boolean = false; set(value) {
        field = value
        hidden = value
    }
    @JvmField
    val info = this::class.findAnnotation<ButtonInfo>() ?: throw Exception("Missing @ButtonInfo annotaton")
    final override var color = info.parent.categoryColor; get() {
        return info.parent.box.color.let {
            if (field == it) field else it
        }
    }
    open val title: String = this@ButtonImpl::class.simpleName.toString()
    protected open fun execAction() = Unit
    private var wasWithinBounds: Boolean = false
    var toggled: Boolean = false; private set(value) {
        if (field xor value) {
            if (value) lightUp() else lightDown()
            field = value
            execAction()
        } else field = value
    }
    private fun lightUp() { opacity += ButtonInfo.lightningFactor }
    private fun lightDown() { opacity -= ButtonInfo.lightningFactor }
    final override fun mouseClicked(d: Double, e: Double, i: Int): Boolean {
        wasWithinBounds = isWithinBounds(d, e).also {
            if (it) {
                if (info.kind == Type.TOGGLE) {
                    toggled = !toggled
                } else if (info.kind == Type.CLICK) lightUp()
            }
        }
        return false
    }

    final override fun mouseReleased(d: Double, e: Double, i: Int): Boolean {
        if (wasWithinBounds && (info.kind == Type.CLICK)) {
            execAction()
            lightDown()
        }; return false
    }
    final override fun render(poseStack: PoseStack?) {
        prepareDraw()
        colorShader(); defaultRectFunc()
        endDraw()
        if (poseStack != null) {
            FontUtil.drawScaled(poseStack, title, x + 1, y + 1, 0.75F)
        }
    }
}
